/* hash.c */
/* Last modified: April 1, 2007 */


/* hash entries (2 * 64 bits), counting bits from right to left
   bits        variable        description
   0-63        key             pos.key
   64-79       score           actual score + infinity (it's always positive)
   80-87       depth
   88-93       from            hash move
   94-99       to              hash move
   100         mateThreat      dangerous situation after null move 
   101-103     hashFlag        EXACT, FAIL_LOW, or FAIL_HIGH
*/   


void hashStore(ui wToMove, ui ply, ui depth, int score, ui from, ui to, 
	       ui mateThreat, ui hashFlag, sPlyData *thisPlyData, sHash **hash,
	       sPos *pos)
{
  struct HashTable *thisHash;
  ui index;

  ASSERT(wToMove == false || wToMove == true);
  ASSERT(ply >= 0 && ply < maxPly);
  ASSERT(depth >= 0 && depth < maxPly);
  ASSERT(score >= -mateScore && score <= mateScore);
  ASSERT(from >= A1 && from <= H8);
  ASSERT(to >= A1 && to <= H8);
  ASSERT(mateThreat == false || mateThreat == true);
  ASSERT(hashFlag == EXACT || hashFlag == FAIL_LOW || hashFlag == FAIL_HIGH);
  ASSERT(thisPlyData != NULL);
  ASSERT(*hash != NULL);
  ASSERT(pos != NULL);

  /* locate the storage area */
  ASSERT(pos->hashSize > 0);
  index = (ui)(thisPlyData->key & pos->hashSize);
  ASSERT(index <= pos->hashSize);
  if (wToMove) thisHash = &hash[white][index];
  else thisHash = &hash[black][index];

  /* deep enough?  then store new data */
  if (depth >= ((ui)(thisHash->stuffs >> 16) & 0xFF))
    {
      /* back up first, */
      if (((ui)(thisHash->stuffsSave >> 16) & 0xFF) <
	  ((ui)(thisHash->stuffs >> 16) & 0xFF))
	{
	  thisHash->keySave = thisHash->key;
	  thisHash->stuffsSave = thisHash->stuffs;
	}

      /* then replace */
      thisHash->key = thisPlyData->key;
      if (score > almostScore) score += ply;
      else if (score < -almostScore) score -= ply;
      /* *score can be out of bound here due to null move */
      thisHash->stuffs = (((ui)(score + infinity)) & 0xFFFF) | 
	(depth << 16) | (from << 24) | ((BB) to << 30) | 
	((BB) mateThreat << 36) | ((BB) hashFlag << 37);
    }
}


int hashProbe(ui wToMove, ui ply, ui depth, int *score, ui *from1, ui *to1, 
	      ui *mateThreat, ui *hashFlag, 
	      sPlyData *thisPlyData, sHash **hash, sPos *pos)
{
  struct HashTable *thisHash;
  ui index, hashDepth;

  ASSERT(wToMove == false || wToMove == true);
  ASSERT(ply >= 0 && ply < maxPly);
  ASSERT(depth >= 0 && depth < maxPly);
  ASSERT(*from1 == 0);
  ASSERT(*to1 == 0);
  ASSERT(mateThreat != NULL);
  ASSERT(hashFlag != NULL);
  ASSERT(thisPlyData != NULL);
  ASSERT(*hash != NULL);
  ASSERT(pos != NULL);
  ASSERT(pos->hashSize > 0);

  /* locate the storage area */
  ASSERT(pos->hashSize > 0);
  index = (ui)(thisPlyData->key & pos->hashSize);
  ASSERT(index <= pos->hashSize);
  if (wToMove) thisHash = &hash[white][index];
  else thisHash = &hash[black][index];

  /* the key matches and the stored data was from a deeper search */

  /* try the primary storage */
  if (thisHash->key == thisPlyData->key)
    {
      /* get hash move, don't have to be deep enough */
      *from1 = (ui)(thisHash->stuffs >> 24) & 63;
      *to1 = (ui)(thisHash->stuffs >> 30) & 63;
      
      hashDepth = ((ui)(thisHash->stuffs >> 16) & 0xFF);
      ASSERT(hashDepth > 0 && hashDepth <= maxPly);
      *hashFlag = (ui)(thisHash->stuffs >> 37) & 7;
      ASSERT(*hashFlag == FAIL_LOW || *hashFlag == FAIL_HIGH || 
	     *hashFlag == EXACT);
      if (depth <= hashDepth)
	{
	  *mateThreat = (ui)(thisHash->stuffs >> 36) & 1;
	  *score = (thisHash->stuffs & 0xFFFF) - infinity;
	  /* *score can be out of bound here due to null move */
	  if (*score > almostScore) *score -= ply;
	  else if (*score < -almostScore) *score += ply;
	  return true;
	}
    }

  /* try the back up storage */
  if (thisHash->keySave == thisPlyData->key)
    {
      /* get hash move, don't have to be deep enough */
      if (*from1 == 0 && *to1 == 0)
	{
	  *from1 = (ui)(thisHash->stuffsSave >> 24) & 63;
	  *to1 = (ui)(thisHash->stuffsSave >> 30) & 63;
	}
      
      hashDepth = ((ui)(thisHash->stuffsSave >> 16) & 0xFF);
      ASSERT(hashDepth > 0 && hashDepth <= maxPly);
      *hashFlag = (ui)(thisHash->stuffs >> 37) & 7;
      ASSERT(*hashFlag == FAIL_LOW || *hashFlag == FAIL_HIGH || 
	     *hashFlag == EXACT);
      if (depth <= hashDepth)
	{
	  *score = (thisHash->stuffsSave & 0xFFFF) - infinity;
	  /* *score can be out of bound here due to null move */
	  *mateThreat = (ui)(thisHash->stuffsSave >> 36) & 1;
	  if (*score > almostScore) *score -= ply;
	  else if (*score < -almostScore) *score += ply;
	  return true;
	}
    }
  
  return false;
}
