/* search.c */
/* Last modified: July 15, 2007 */


int qsearch(ui wToMove, ui ply, ui qDepth, int alpha, int beta,  
	    sPos *pos, sPlyData *plyData, sPV *pv, sOthers *others, sTC *tc, 
	    ui nullMode, ui nodeType)
{
#ifdef VERIFY
  char s[8];
#endif
  BB winningCapture[32];

  BB theMove;
  ui i, j, noOfCaptures, notOK;
  int score = 0;
  int wPieceScore, bPieceScore, lScore, doubleBonus = 0;
  ui movesSearched, nodeType1;
#ifdef WINDOWS
  double clock1;
#else
  struct timeval tv;
  struct timezone tz;
  BB clock1;
#endif

  ASSERT(wToMove == false || wToMove == true);
  ASSERT(ply >= 0 && ply <= maxPly);
  ASSERT(qDepth >= 0);
  ASSERT(alpha < beta);
  ASSERT(alpha >= -mateScore && alpha <= mateScore);
  ASSERT(beta >= -mateScore && beta <= mateScore);
  ASSERT(pos != NULL);
  ASSERT(plyData != NULL);
  ASSERT(pv != NULL);
  ASSERT(others != NULL);
  ASSERT(tc != NULL);
  ASSERT(!(pos->WBoard & pos->BBoard));
  ASSERT(nullMode == false || nullMode == true);
  ASSERT(nodeType == allNode || nodeType == PVNode || nodeType == cutNode);

  pv[ply].length = ply;

  /* too deep */
  if (ply >= maxPly-1) 
    {
      pos->Flags |= OneMask[4]; 
      return alpha;
    }

#ifdef _REENTRANT
  /* time to end pondering */
  if (tc->timeBudget > maxSearchBudget)
    {
      ASSERT(go != goSearch);
      if (go != goPonder)
	{
	  pos->Flags |= OneMask[4];    /* terminate the search */
	  return 0;
	}
    }
  /* Time check */
  else 
#endif
    if ((pos->sNodes + pos->qNodes) > tc->nodeCheckPoint && 
      pos->iteration > 1) 
    {
      ASSERT(go != goPonder);
#ifdef WINDOWS
      clock1 = 1000000.0/CLOCKS_PER_SEC*clock();
#else
      gettimeofday(&tv, &tz);
      clock1 = (BB) tv.tv_sec*1000000 + (BB) tv.tv_usec;
#endif

      ASSERT(clock1 > tc->clock0);
      if ((double) (clock1-tc->clock0) >= tc->timeBudget)
	{
	  pos->Flags |= OneMask[4];    /* terminate the search */
	  return 0;
	}
      else if (!(pos->Flags & OneMask[7]) &&
	       (double) (clock1-tc->clock0) >= 0.9*tc->timeBudget)
	{
	  pos->Flags |= OneMask[7];    /* prepare to terminate the search */
	  tc->nodeInc /= 10;              /* check more often now */
	}

      tc->nodeCheckPoint += tc->nodeInc;
    }

  pos->qNodes++;

#ifdef TREEDUMP
  printf("TREEDUMP at beginning of qsearch(): nodes=%.0f ply=%d key=0x%016llX\n",
	 (double)(pos->sNodes + pos->qNodes), ply, plyData[ply].key);
  display(pos);
#endif

  /* Are we in check?  In qsearch(), we don't care. */

  /* evaluate the position */
  /* IMPORTANT: evalMaterial() returns white's lScore */
  lScore = evalMaterial(pos, others, &wPieceScore, &bPieceScore, &doubleBonus);
  ASSERT(lScore >= -mateScore && lScore <= mateScore);
  if (lScore == drawScore) 
    {
      if (0 >= beta)
	{
#ifdef TREEDUMP
	  printf("%d:0>=%d (fail high with drawScore in qsearch())\n", 
		 wToMove, beta);
#endif
	  return beta;
	}
      if (0 > alpha) return 0;
      return alpha;
    }

  /* try to save time when pos->iteration <= pos->fastIteration */
  if (pos->iteration <= minPly || pos->iteration > pos->fastIteration)
    score = eval(wToMove, ply, pos, plyData, others, lScore,
		 wPieceScore, bPieceScore, alpha, doubleBonus);
  else score = lScore;
  ASSERT(score >= -mateScore && score <= mateScore);

  /* now assign score to the right color */
  if (wToMove) score = pos->wScore + score;
  else score = -pos->wScore -score;
#ifdef TREEDUMP
	  printf("%d's eval in qsearch(): %d\n", wToMove, score);
#endif

  /* stand pat, anyone? */
  if (score > alpha)
    {
      if (score >= beta)
	{
#ifdef TREEDUMP
	  printf("%d:%d>=%d (fail high after eval() in qsearch())\n", 
		 wToMove, score, beta);
#endif
	  return beta;
	}
      alpha = score;
    }

  /* try captures */
  noOfCaptures = qcaptures(wToMove, ply, winningCapture, pos, plyData, others);

  /* can't go any deeper */
  if (noOfCaptures == 0)
    {
#ifdef TREEDUMP
      printf("%d:%d (no move in qsearch())\n", wToMove, alpha);
#endif
      return alpha;
    }

  /* go deeper */
  movesSearched = 0;
  for (i = 0; i < noOfCaptures; i++)
    {
      pick(i, noOfCaptures, winningCapture);
      theMove = winningCapture[i];
      ASSERT(from(theMove) <= H8 && to(theMove) <= H8 &&
	     from(theMove) != to(theMove));
      ASSERT(piece(theMove) < 13);

#ifdef VERIFY
	  algebraic(theMove, s);
	  sprintf(pos->posID, "%d._%.0f_%d_0x%016llX_%d_%s", 
		  pos->moveNumber,
		  (double)(pos->sNodes + pos->qNodes),
		  ply, plyData[ply].key, wToMove, s);
#ifdef TREEDUMP
	  printf("TREEDUMP in qsearch(): %s ", pos->posID);
	  printf("%d<%d<%d\n", alpha, score, beta);
#endif
#endif
      makeMove(wToMove, ply+1, theMove, pos, plyData, others);
      if (wToMove) 
	notOK = (getBAttacker(pos->wKSq, pos, &plyData[ply+1], others) & 2);
      else
	notOK = (getWAttacker(pos->bKSq, pos, &plyData[ply+1], others) & 2);
      if (notOK)
	{
	  unmakeMove(wToMove, ply+1, theMove, pos, plyData);
	  continue;
	}
      else
	{
	  movesSearched++;

	  ASSERT(ply < maxPly-1);
	  if (nodeType != PVNode || movesSearched == 1)
	    {
	      if (movesSearched == 1) nodeType1 = -nodeType;
	      else nodeType1 = allNode;
	      ASSERT(nodeType1 == allNode || nodeType1 == PVNode || 
		     nodeType1 == cutNode);
	      score = -qsearch(wToMove^1, ply+1, qDepth+1, -beta,
			       -alpha, pos, plyData, pv, others, tc, 
			       nullMode, nodeType1);
	      ASSERT(score >= -mateScore && score <= mateScore);
	    }
	  else
	    {
	      ASSERT(nodeType == PVNode);
	      score = -qsearch(wToMove^1, ply+1, qDepth+1, 
			       -alpha-1, -alpha, pos, plyData, pv, 
			       others, tc, nullMode, cutNode);
	      ASSERT(score >= -mateScore && score <= mateScore);
	      /* re-search, keep the nodeType */
	      if (score > alpha /* && score < beta */)
		{
#ifdef TREEDUMP
		  printf("score=%d > alpha=%d -> re-qsearch in qsearch()\n",
			 score, alpha);
#endif
		  score = -qsearch(wToMove^1, ply+1, qDepth+1, 
				   -beta, -alpha, pos, plyData, pv, 
				   others, tc, nullMode, nodeType);
		  ASSERT(score >= -mateScore && score <= mateScore);
		}
	    }
	  if (nodeType == cutNode) nodeType = allNode;

	  if (pos->Flags & OneMask[4]) 
	    {
	      unmakeMove(wToMove, ply+1, theMove, pos, plyData);
	      return 0;
	    }
	  
	  if (score > alpha)
	    {
	      if (score >= beta) 
		{
		  unmakeMove(wToMove, ply+1, theMove, pos, plyData);
#ifdef TREEDUMP
		  printf("%d:%d>=%d (fail high in qsearch())\n", 
			 wToMove, score, beta);
#endif
		  /* no need to store the score here, zero depth is useless */
		  return beta;
		}
	      
	      alpha = score;

	      /* update the PV */
	      if (!nullMode)
		{
		  pv[ply].move[ply] = theMove;
		  ASSERT(pv[ply+1].length > ply &&
			 pv[ply+1].length <= maxPly)
		  for (j = ply+1; j < pv[ply+1].length; j++)
		    pv[ply].move[j] = pv[ply+1].move[j];
		  /* pv[ply].length = pv[ply+1].length; */
		}
	    }
	  unmakeMove(wToMove, ply+1, theMove, pos, plyData);
	}
    }
  
#ifdef TREEDUMP
  printf("%d:%d (fail low in qsearch())\n", wToMove, alpha);
#endif

  return alpha;
}


static int search(ui wToMove, ui ply, int alpha, int beta, ui depth, sPos *pos,
		  sPlyData *plyData, sPV *pv, sHash **hash, sOthers *others, 
		  sTC *tc, ui nullOK, ui nullMode, int centiExtension, 
		  ui nodeType, ui followPV, sPV *thePV, BB lastMove)
{
  BB winningCapture[32], nonCapture[128], allMoves[128];

  BB theMove, bestMove = (BB) 0;
  BB Checkers;
  ui i, j, noOfCaptures = 0, noOfNonCaptures =  0, nMoves = 0, notOK;
  int score = 0;

  ui movesSearched, searchMode, nodeType1;
  ui from1 = 0, to1 = 0, p, f, t;

  ui hashFlag = FAIL_LOW;
  ui checkers;
  int origCE;
  int extension = 0;
  ui nullReduction, mateThreat = 0, pieces, newDepth;
#ifdef WINDOWS
  double clock1;
#else
  struct timeval tv;
  struct timezone tz;
  BB clock1;
#endif
#ifdef VERIFY
  char s[8];
#endif

  ASSERT(wToMove == false || wToMove == true);
  ASSERT(ply >= 0 && ply < maxPly);
  ASSERT(alpha < beta);
  ASSERT(alpha >= -mateScore && alpha <= mateScore);
  ASSERT(beta >= -mateScore && beta <= mateScore);
  ASSERT(depth >= 1 && depth < maxPly);
  ASSERT(pos != NULL);
  ASSERT(plyData != NULL);
  ASSERT(*hash != NULL);
  ASSERT(pv != NULL);
  ASSERT(others != NULL);
  ASSERT(tc != NULL);
  ASSERT(nullOK == false || nullOK == true);
  ASSERT(nullMode == false || nullMode == true);
  ASSERT(nodeType == allNode || nodeType == PVNode || nodeType == cutNode);
  ASSERT(!(pos->WBoard & pos->BBoard));
  ASSERT(followPV >= 0 && followPV <= 3);

  pv[ply].length = ply;

  /* over quota */
  if (ply >= maxPly-1) 
    {
      pos->Flags |= OneMask[4];
      return alpha;
    }

#ifdef _REENTRANT
  /* time to end pondering */
  if (tc->timeBudget > maxSearchBudget)
    {
      ASSERT(go != goSearch);
      if (go != goPonder)
	{
	  pos->Flags |= OneMask[4];    /* terminate the search */
	  return 0;
	}
    }
  /* Time check */
  else 
#endif
    if ((pos->sNodes + pos->qNodes) > tc->nodeCheckPoint && 
      pos->iteration > 1) 
    {
      ASSERT(go != goPonder);
#ifdef WINDOWS
      clock1 = 1000000.0/CLOCKS_PER_SEC*clock();
#else
      gettimeofday(&tv, &tz);
      clock1 = (BB) tv.tv_sec*1000000 + (BB) tv.tv_usec;
#endif

      ASSERT(clock1 > tc->clock0);
      if ((double) (clock1-tc->clock0) >= tc->timeBudget)
	{
	  pos->Flags |= OneMask[4];    /* terminate the search */
	  return 0;
	}
      else if (!(pos->Flags & OneMask[7]) &&
	       (double) (clock1-tc->clock0) >= 0.9*tc->timeBudget)
	{
	  pos->Flags |= OneMask[7];    /* prepare to terminate the search */
	  tc->nodeInc /= 10;              /* check more often now */
	}

      tc->nodeCheckPoint += tc->nodeInc;
    }

  pos->sNodes++;

#ifdef TREEDUMP
  printf("TREEDUMP at beginning of search(): nodes=%.0f ply=%d key=0x%016llX\n",
	 (double)(pos->sNodes + pos->qNodes), ply, plyData[ply].key);
  display(pos);
#endif

  /* probing Nalimov's endgame table base */
  if (ply <= pos->iteration && ply < minPly &&
      bitCount(pos->RotatedBoard90) <= pos->EGTBlimit)
    {
      if (EGTBProbe(wToMove, &score, pos, &plyData[ply])) 
	{
#ifdef TREEDUMP
	  printf("%d:%d EGTBProbe() result in search())\n", 
		 wToMove, score);
#endif
	  ASSERT(score >= -mateScore && score <= mateScore);
	  return score;
	}
    }
 
  /* see if we can skip the search */
  if (hashProbe(wToMove, ply, depth, &score, &from1, &to1, 
		&mateThreat, &hashFlag, &plyData[ply], hash, pos))
    {
#ifdef TREEDUMP
      printf("hashProbe'd score=%d is ", score);	      
#endif
      if (nodeType != PVNode)    /* if PVNode use the hash moves only */
	{
	  switch(hashFlag)
	    {
	    case EXACT:
#ifdef TREEDUMP
	      printf("EXACT\n");	      
#endif
	      /* score can be < -mateScore or > mateScore here.  Just let the
		 alpha-beta mechanism in the previous ply take care of it. */
	      return score;
	    case FAIL_LOW:
#ifdef TREEDUMP
	      printf("FAIL_LOW.  alpha=%d\n", alpha);	      
#endif
	      if (score <= alpha) return alpha;
	      break;
	    case FAIL_HIGH:
#ifdef TREEDUMP
	      printf("FAIL_HIGH.  beta=%d\n", beta);	      
#endif
	      if (score >= beta) return beta;
	      break;
	    default:
	      printf("ERROR: undefined hashFlag in search()\n");
	      break;
	    }
	}
    }


  /* Are we being checked? */
  Checkers = (BB) 0;
  if (wToMove) 
    checkers = isInCheckByB(pos->wKSq, &Checkers, pos, others);
  else checkers = isInCheckByW(pos->bKSq, &Checkers, pos, others);

  /* null move */
  if (wToMove) pieces = bitCount(pos->WBoard);
  else pieces = bitCount(pos->BBoard);
  if (nodeType != PVNode &&
      !checkers &&                /* side to move is not in check */
      nullOK &&                   /* previous move was not a null move */
      /* side to move have enough pieces to move */
      (pieces > 5 /* 9 || depth < 7 */))
    {
#ifdef TREEDUMP
      printf("NULL move\n");
#endif
      plyData[ply+1] = plyData[ply];
      if (plyData[ply+1].EP) 
	plyData[ply+1].key ^= others->ePKey[firstOne(plyData[ply+1].EP)];
      plyData[ply+1].count50Moves++;
      plyData[ply+1].EP = 0;
      if (depth > 6 && pieces > 9) nullReduction = 3 + (pos->iteration >> 3);
      else nullReduction = 3;

      if (depth > (nullReduction+1))
	{
	  score = -search(wToMove^1, ply+1, -beta, -beta+1, 
			  depth-nullReduction-1, pos, plyData, pv, hash, 
			  others, tc, 0, 1, centiExtension, -nodeType, false,
			  thePV, (BB)0);
	  ASSERT(score >= -mateScore && score <= mateScore);
	}
      else
	{
	  score = -qsearch(wToMove^1, ply+1, 0, -beta, -beta+1, 
			   pos, plyData, pv, others, tc, 1, -nodeType);
	  ASSERT(score >= -mateScore && score <= mateScore);
	}
      if (pos->Flags & OneMask[4]) return 0;
      if (score >= beta) 
	{
#ifdef TREEDUMP
	  printf("%d:%d>=%d (fail high in null move())\n", 
		 wToMove, score, beta);
#endif
	  return beta;
	}
      /* A dangerous situation: if we don't move, we get mated - 
	 Bruce Moreland */
      if (score < -almostScore) mateThreat = 1;
    }


  /* Generate moves */
  if (checkers)
    {
      if (wToMove)
	  nMoves = outOfBCheck(ply, pos->wKSq, Checkers, allMoves, 
			       pos, plyData, others);
      else
	  nMoves = outOfWCheck(ply, pos->bKSq, Checkers, allMoves, 
			       pos, plyData, others);
#ifdef VERIFY
      noOfCaptures = captures(wToMove, ply, winningCapture,
			      pos, plyData, others, 0);
      noOfNonCaptures = nonCaptures(wToMove, ply, nonCapture, pos,
				    plyData, others, 0);
      if (noOfCaptures+noOfNonCaptures != nMoves)
	{
	  ui x;

	  if (wToMove) x = pos->wKSq;
	  else x = pos->bKSq;

	  display(pos);
	  for (j = 0; j < noOfCaptures; j++) 
	    printf("ERROR in search C%d. %d-%d\n", j+1, 
		   from(winningCapture[j]), to(winningCapture[j]));
	  for (j = 0; j < noOfNonCaptures; j++) 
	    printf("ERROR in search N%d. %d-%d\n", j+1, 
		   from(nonCapture[j]), to(nonCapture[j]));
	  for (j = 0; j < nMoves; j++) 
	    {
	    printf("ERROR in search A%d. %d-%d\n", j+1, 
		   from(allMoves[j]), to(allMoves[j]));
	    printf("direction from my king to moving piece is %d\n", 
		   others->direction[x][from(allMoves[j])]);
	    }
	  exit(EXIT_FAILURE);
	}
#endif

      if (nMoves == 0) 
	{
#ifdef TREEDUMP
	  printf("%d: checkmated in %d moves.\n", wToMove, ply);
#endif
	  return (ply -1 -mateScore);
	}
    }
  else
    {
      /* try captures */
      noOfCaptures = captures(wToMove, ply, winningCapture,
			      pos, plyData, others, 1);
      /* try non-captures */
      noOfNonCaptures = nonCaptures(wToMove, ply, nonCapture, pos,
				    plyData, others, 1);

      /* Copying all moves to allMoves[]. */
      nMoves = 0;
      if (noOfCaptures > 0)
	for (i = 0; i < noOfCaptures; i++)
	  {
	    allMoves[nMoves++] = winningCapture[i];
	  }
      if (noOfNonCaptures > 0)
	for (i = 0; i < noOfNonCaptures; i++)
	  {
	    allMoves[nMoves++] = nonCapture[i];
	  }
    }


  /* Trust PV move better than have hash move.       */
  /* Unfortunately, PV move can't go pass null move. */
  if ((followPV & 1) && thePV->length >= ply)
    {
      from1 = from(thePV->move[ply]);
      to1 = to(thePV->move[ply]);
      ASSERT(from1 >= A1 && from1 <= H8);
      ASSERT(to1 >= A1 && to1 <= H8);
    }

  /* No hash move, no PV move, let's do Internal Iterative Deepening */
  if (from1 == to1 && depth >= iidDepth)
    {
      if (((ply & 1) && alpha == pos->rootAlpha && beta == pos->rootBeta) ||
	  (!(ply & 1) && alpha == pos->rootAlpha && beta == pos->rootBeta))
	{
	  newDepth = depth-iidReduction;
	  ASSERT(newDepth > 0);
	  if (newDepth > 1)
	    {
	      score = search(wToMove, ply, alpha, beta, newDepth-1, pos, 
			     plyData, pv, hash, others, tc, true, nullMode, 
			     centiExtension, nodeType, followPV, thePV, (BB)0);
	      if (score <= alpha)
		score = search(wToMove, ply, -mateScore, beta, newDepth-1, pos,
			       plyData, pv, hash, others, tc, true, nullMode, 
			       centiExtension, nodeType, followPV, thePV, 
			       (BB)0);
	    }
	  else
	    {
	      score = qsearch(wToMove, ply, 0, alpha, beta, pos, 
			      plyData, pv, others, tc, nullMode, nodeType);
	      if (score <= alpha)
		score = qsearch(wToMove, ply, 0, -mateScore, alpha, 
				pos, plyData, pv, others, tc, nullMode, 
				nodeType);
	    }
	  from1 = from(pv[ply].move[ply]);
	  to1 = to(pv[ply].move[ply]);
	}
    }


  /* mark the moves as hash move or killer move */
  if (from1 != to1)
    {
      ui breaker = 2;
      for (i = 0; i < nMoves && breaker; i++)
	{
	  f = from(allMoves[i]);
	  t = to(allMoves[i]);

	  /* give onlyScore to hash move #1 */
	  if (f == from1 && t == to1)
	    {
	      /* For some reason, doing hash move first works best */
	      allMoves[i] = (allMoves[i] & ~ScoreBoard) |
		((BB)(onlyScore+mateScore) << 30);
	      breaker--;
	    }
	  /* give 50 points to killer so it will come after winning captures */
	  else if (f == pos->killerFrom[ply] &&
		   t == pos->killerTo[ply])
	    {
	      if (score(allMoves[i]) <= 1)
		allMoves[i] = (allMoves[i] & ~ScoreBoard) |
		  ((BB)(50+mateScore) << 30);
	      breaker--;
	    }
	}
    }
  else
    {
      for (i = 0; i < nMoves; i++)
	{
	  t = to(allMoves[i]);
	  /* give 50 points to killer so it will come after winning captures */
	  if (from(allMoves[i]) == pos->killerFrom[ply] &&
	      t == pos->killerTo[ply])
	    {
	      if (score(allMoves[i]) <= 1)
		allMoves[i] = (allMoves[i] & ~ScoreBoard) |
		  ((BB)(50+mateScore) << 30);
	      break;
	    }
	}
    }

  /* First, be careful not to use the junk score from hashProbe() or
     null move.
   */
  score = alpha;
  hashFlag = FAIL_LOW;        /* initialize hashFlag */
  movesSearched = 0;
  origCE = centiExtension;
  bestMove = (BB) 0;

  /* main loop in search() */
  for (i = 0; i < nMoves; i++)
    {
      pick(i, nMoves, allMoves);
      theMove = allMoves[i];
      p = piece(theMove);
      f = from(theMove);
      t = to(theMove);
      ASSERT(f <= H8 && t <= H8 && f != t);
      ASSERT(p < 13);

#ifdef VERIFY
      algebraic(theMove, s);
      sprintf(pos->posID, "%d._%.0f_%d_0x%016llX_%d_%s", 
	      pos->moveNumber, (double)(pos->sNodes + pos->qNodes),
	      ply, plyData[ply].key, wToMove, s);
#ifdef TREEDUMP
      printf("TREEDUMP in search(): %s ", &pos->posID);
      printf("%d<%d<%d\n", alpha, score, beta);
#endif
#endif
      makeMove(wToMove, ply+1, theMove, pos, plyData, others);
      notOK = false;
      if (!checkers)
	{
	  if (wToMove) 
	    notOK = (getBAttacker(pos->wKSq, pos, &plyData[ply+1], others) & 
		     2);
	  else
	    notOK = (getWAttacker(pos->bKSq, pos, &plyData[ply+1], others) & 
		     2);
	}

      if (notOK)
	{
	  unmakeMove(wToMove, ply+1, theMove, pos, plyData);
	  continue;
	}
      else
	{
	  movesSearched++;

	  /* Is this position a draw by rule? */
	  searchMode = 1;
	  if (plyData[ply+1].count50Moves > 2) 
	    switch(isDrawByRule(ply+1, pos, &plyData[ply+1]))
	      {
	      case 2:
		searchMode = 2;
		break;
	      case 3:
		score = 0;
		searchMode = 0;
		break;
	      case 13:
		score = 0;
		searchMode = 0;
		break;
	      case 50:
		score = 0;
		searchMode = 0;
		break;
	      default:
		break;
	      }
	  
	  if (searchMode)
	    {
	      /* Don't forget to initialize these variables. */
	      centiExtension = extension = 0;

	      if (nodeType != allNode)
		{
		  /* Following PV */
		  if ((followPV & 1) &&
		      movesSearched == 1 &&
		      f == from1 &&
		      t == to1)
		    {
		      nodeType = PVNode;
		      centiExtension += pos->pvExt;
		    }
		  else followPV &= 2;        /* unset bit #1 */
		  
		  if (checkers) 	      
		    {
		      /* check entension */
		      centiExtension += pos->checkExt;
		      /* only one reply available, extend */
		      if (nMoves == 1) centiExtension += pos->onlyReplyExt;
		    }
		  
		  /* if in a dangerous situation, extend */
		  if (mateThreat) centiExtension += pos->mateThreatExt;
		  
		  /* parrot extension */
		  if (t == to(lastMove) && score(theMove) > 1 && 
		      lastMove != (BB)0)
		    centiExtension += pos->parrotExt;
		  
		  /* if pawn moves to the 7th rank, extend */
		  if (row(t) == 6 && p == wP) 
		    centiExtension += pos->pawn7thExt;
		  else if (row(t) == 1 && p == bP) 
		    centiExtension += pos->pawn7thExt;
		  
		  /* based on number of captures, extend */
		  /* centiExtension += (noOfCaptures*5); */
		  
		  /* new 7th rank rook or queen condition? extend
		     if (seventh) centiExtension += 75; */
		  
		  /* reduce extension after 2*pos->iteration */
		  if (ply > 2*pos->iteration)
		    {
		      /* integer operation, it is not slow */
		      if (ply <= 4*pos->iteration)
			centiExtension = centiExtension *
			  (4*pos->iteration-ply) / (2*pos->iteration);
		      else centiExtension = 0;
		    }
		}
	      
	      /* add previous ply's centiExtension */ 
	      centiExtension += origCE;
	      
	      /* cash in the partial ply extensions */
	      if (centiExtension >= 100)
		{
		  centiExtension -= 100;
		  extension = 1;
		}
	      
	      ASSERT(depth > 0);
	      /* late move reduction */
	      if (pos->rootMaxPieceScore >= endGameValue &&
		  /* nodeType == allNode && */
		  extension == 0 && centiExtension == origCE && 
		  score(theMove) <= 1 && !checkers && 
		  movesSearched > lateMoveThreshold)
 		{
		  ASSERT(noOfCaptures <= nMoves);
		  if (depth > 2) 
		    {
		      if (pos->iteration > minPly &&
			  pos->iteration <= pos->fastIteration &&
			  wKPenalty(ply, pos, plyData, others) > -pValue &&
			  bKPenalty(ply, pos, plyData, others) < pValue)
			extension = -depth+1;  /* fall straight to qsearch() */
		      else extension = -1;
		    }
		  else if (depth == 2 && 
			   wKPenalty(ply, pos, plyData, others) > -pValue &&
			   bKPenalty(ply, pos, plyData, others) < pValue)
		    extension = -1;
		}

	      
	      ASSERT(ply >= 0 && ply < maxPly-1);
	      if (nodeType != PVNode || movesSearched == 1)
		{
		  if (movesSearched == 1) nodeType1 = -nodeType;
		  else nodeType1 = allNode;
		  ASSERT(nodeType1 == allNode || nodeType1 == PVNode || 
			 nodeType1 == cutNode);
		  if (depth+extension > 1)
		    {
		      score = -search(wToMove^1, ply+1, -beta, -alpha, 
				      depth+extension-1, pos, plyData, pv,
				      hash, others, tc, true, nullMode,
				      centiExtension, nodeType1, followPV,
				      thePV, theMove);
		      ASSERT(score >= -mateScore && score <= mateScore);
		      /* re-search, keep the nodeType */
		      if (extension < 0 && score > alpha)
			{
			  score = -search(wToMove^1, ply+1, -beta, -alpha,
					  depth-1, pos, plyData, pv, hash,
					  others, tc, true, nullMode, 
					  centiExtension, nodeType,
					  followPV, thePV, theMove);
			}
		    }
		  else
		    {
		      score = -qsearch(wToMove^1, ply+1, 0, -beta, -alpha,
				       pos, plyData, pv, others, tc, 
				       nullMode, nodeType1);
		      ASSERT(score >= -mateScore && score <= mateScore);
		      /* re-search, keep the nodeType */
		      if (extension < 0 && score > alpha)
			{
			  score = -search(wToMove^1, ply+1, -beta, -alpha,
					  depth-1, pos, plyData, pv, hash,
					  others, tc, true, nullMode, 
					  centiExtension, nodeType,
					  followPV, thePV, theMove);
			}
		    }
		}
	      else
		{
		  ASSERT(nodeType == PVNode);
		  if (depth+extension > 1)
		    {
		      score = -search(wToMove^1, ply+1, -alpha-1, -alpha, 
				      depth+extension-1, pos, plyData, pv,
				      hash, others, tc, true, nullMode,
				      centiExtension, cutNode, false, 
				      thePV, theMove);
		      ASSERT(score >= -mateScore && score <= mateScore);
		      /* re-search, keep the nodeType */
		      if (score > alpha)
			{
#ifdef TREEDUMP
			  printf("score=%d > alpha=%d -> re-search in search()\n",
				 score, alpha);
#endif
			  score = -search(wToMove^1, ply+1, -beta, -alpha, 
					  depth+extension-1, pos, plyData, 
					  pv, hash, others, tc, true, 
					  nullMode, centiExtension, 
					  nodeType, false, thePV, theMove);
			  ASSERT(score >= -mateScore && 
				 score <= mateScore);
			}
		    }
		  else /* not deep enough */
		    {
		      score = -qsearch(wToMove^1, ply+1, 0, -alpha-1, 
				       -alpha, pos, plyData, pv,
				       others, tc, nullMode, cutNode);
		      ASSERT(score >= -mateScore && score <= mateScore);
		      /* re-search, keep the nodeType */
		      if (score > alpha)
			{
#ifdef TREEDUMP
			  printf("score=%d > alpha=%d -> re-qsearch in search()\n", 
				 score, alpha);
#endif
			  score = -qsearch(wToMove^1, ply+1, 0, -beta, 
					   -alpha, pos, plyData, pv,
					   others, tc, nullMode, nodeType);
			  ASSERT(score >= -mateScore && 
				 score <= mateScore);
			}
		    }
		}

	      if (pos->Flags & OneMask[4]) 
		{
		  unmakeMove(wToMove, ply+1, theMove, pos, plyData);
		  return 0;
		}
	      if (searchMode == 2) score = score / 2;
	    }
	  if (nodeType == cutNode) nodeType = allNode;

	  if (score > alpha)
	    {
	      if (score >= beta)
		{
		  unmakeMove(wToMove, ply+1, theMove, pos, plyData);
#ifdef TREEDUMP
		  printf("%d:%d>=%d (fail high in search())\n", 
			 wToMove, score, beta);
#endif
                 /* killer move */
		  if (score(theMove) <= 1)   /* include even/losing capture */
                   {
                     pos->killerFrom[ply] = f;
                     pos->killerTo[ply] = t;
                   }

		  /* store the score in hash table */
		  hashStore(wToMove, ply, depth, beta, f, t, mateThreat, 
			    FAIL_HIGH, &plyData[ply], hash, pos);
		  
		  return beta;
		}

	      alpha = score;
	      hashFlag = EXACT;
	      /* high branching factor, keep bestMove */
	      if (followPV & 2) bestMove = theMove;

	      /* update the PV */
	      if (!nullMode)
		{
		  pv[ply].move[ply] = theMove;
		  if (!searchMode) pv[ply].length = ply+1;
		  else
		    {
		      ASSERT(pv[ply+1].length >= ply &&
			     pv[ply+1].length <= maxPly)
		      for (j = ply+1; j < pv[ply+1].length; j++)
			pv[ply].move[j] = pv[ply+1].move[j];
		      pv[ply].length = pv[ply+1].length;
		    }
		}
	    }
	  unmakeMove(wToMove, ply+1, theMove, pos, plyData);
	}
    }

#ifdef TREEDUMP
  printf("%d:%d (fail low in search())\n", wToMove, alpha);
#endif

  if (movesSearched) 
    {
      /* store the score in hash table */
      hashStore(wToMove, ply, depth, alpha, from(bestMove), 
		to(bestMove), mateThreat, hashFlag, &plyData[ply], hash, pos);
		  
      return alpha;
    }
  else
    { 
#ifdef TREEDUMP
      printf("%d: stalemate in %d moves.\n", wToMove, ply);
#endif
      /* store the score in hash table */
      hashStore(wToMove, ply, depth, 0, 0, 0,
		mateThreat, EXACT, &plyData[ply], hash, pos);
		  
      return 0;
    }
}


int rsearch(ui wToMove, int alpha, int beta, ui depth, sPos *pos, 
	    sPlyData *plyData, sPV *pv, sHash **hash, sOthers *others, sTC *tc,
	    BB* allMoves, ui nMoves, int origCE, ui highBF, sPV *thePV)
{
  int score = 0;
  int centiExtension, extension;
  ui pieceCnt;
  ui i, j, searchMode, movesSearched;
  BB theMove = (BB) 0, firstMove = (BB) 0;
  ui t;
  ui followPV;
#ifdef VERIFY
  char s[8];
#endif

  ASSERT(wToMove == false || wToMove == true);
  ASSERT(alpha < beta);
  ASSERT(alpha >= -mateScore && alpha <= mateScore);
  ASSERT(beta >= -mateScore && beta <= mateScore);
  ASSERT(depth >= 1 && depth <= maxPly);
  ASSERT(pos != NULL);
  ASSERT(plyData != NULL);
  ASSERT(*hash != NULL);
  ASSERT(pv != NULL);	    
  ASSERT(others != NULL);
  ASSERT(tc != NULL);
  ASSERT(allMoves != NULL);
  ASSERT(nMoves >= 1);
  ASSERT(highBF >= 0);
  ASSERT(!(pos->WBoard & pos->BBoard));

  pos->sNodes++;
  pv[0].length = 0;

#ifdef TREEDUMP
  printf("TREEDUMP at beginning of rsearch(): nodes=%.0f key=0x%016llX\n",  
	 (double)(pos->sNodes + pos->qNodes), plyData[0].key);
  display(pos);
#endif

  /* probing Nalimov's endgame table base */
  if (bitCount(pos->RotatedBoard90) <= pos->EGTBlimit)
    {
      if (EGTBProbe(wToMove, &score, pos, &plyData[0])) depth = 2;
    }

  /* Set scores for move ordering */
  if (pos->iteration == 1)
  for (i = 0; i < nMoves; i++)
    {
      /* For a looong time, I didn't put the PV move on top of the list. */
      if (from(allMoves[i]) == from(thePV->move[0]) &&
          to(allMoves[i]) == to(thePV->move[0]))
        {
          allMoves[i] = (allMoves[i] & ~ScoreBoard) |
            ((BB)(onlyScore+pos->iteration+mateScore) << 30);
        }
      /* give 50 points to killer so it will come after winning captures */
      else if (from(allMoves[i]) == pos->killerFrom[0] &&
          to(allMoves[i]) == pos->killerTo[0])
        {
          if (score(allMoves[i]) <= 1)
            allMoves[i] = (allMoves[i] & ~ScoreBoard) |
              ((BB)(50+mateScore) << 30);
        }
    }

  /* go deeper */
  movesSearched = 0;
  for (i = 0; i < nMoves; i++)
    {
      pick(i, nMoves, allMoves);
      theMove = allMoves[i];
      ASSERT(from(theMove) <= H8 && to(theMove) <= H8 &&
	     from(theMove) != to(theMove));
      ASSERT(piece(theMove) < 13);
#ifdef VERIFY
      algebraic(theMove, s);
      sprintf(pos->posID, "%d._%.0f_%d_0x%016llX_%d_%s", 
	      pos->moveNumber, (double)(pos->sNodes + pos->qNodes), 0,
	      plyData[0].key, wToMove, s);
#ifdef TREEDUMP
      printf("TREEDUMP in rsearch(): %s ", &pos->posID);
      printf("%d<%d<%d\n", alpha, score, beta);
#endif
#endif
      makeMove(wToMove, 1, theMove, pos, plyData, others);
      movesSearched++;
      if (movesSearched == 1) firstMove = theMove;
      pos->wScore = 0;

      /* Is this position a draw by rule? */
      searchMode = 1;
      switch(isDrawByRule(1, pos, &plyData[1]))
	{
	case 2:
	  searchMode = 2;
	  break;
	case 3:
	  theMove |= ((BB)(flagDraw3) << 46);
	  score = 0;
	  searchMode = 0;
	  break;
	case 13:
	  theMove |= ((BB)(flagDraw13) << 46);
	  score = 0;
	  searchMode = 0;
	  break;
	case 50:
	  theMove |= ((BB)(flagDraw3) << 46);
	  score = 0;
	  searchMode = 0;
	  break;
	default:
	  break;
	}
      
      if (searchMode)
	{
	  /* Do root evaluation here to get pos->wScore to be added to
	     eval() in qsearch().  Can be used for planning or pruning. */
	  pieceCnt = bitCount(pos->RotatedBoard90) - 
	    (bitCount(pos->WPBoard | pos->BPBoard) + 2);
	  ASSERT(pieceCnt <= 32);
	  if (pieceCnt > 10) reval(theMove, pos, others);
	  
	  /* Don't forget to initialize these variables. */
	  centiExtension = origCE;	      
	  extension = 0;

          /* Following PV */
          if (movesSearched == 1 &&
              from(theMove) == from(thePV->move[0]) &&
              to(theMove) == to(thePV->move[0]))
            {
	      /* do not follow PV in the end game */
	      /* try high branching factor here */
	      if (highBF)
		{
		  followPV = 3;        /* set bits #1 and #2 */
		  centiExtension += pos->pvExt;
		}
	      else followPV = 0;
            }
          else
            {
	      if (highBF) followPV = 2;        /* set bit #2 only */
              else followPV = 0;

	      /* Check and only one reply out of check extensions had been
		 added in main(). */

	      /* if pawn moves to the 7th rank, extend */
	      t = to(theMove);
	      if (row(t) == 6 && piece(theMove) == wP) 
		centiExtension += pos->pawn7thExt;
	      else if (row(t) == 1 && piece(theMove) == bP)
		centiExtension += pos->pawn7thExt;
	      if (centiExtension < 100) extension = 0;
	      else        /* cash in the partial extension */
		{
		  centiExtension -= 100;
		  extension = 1;
		}
	    }
	  
	  if (depth+extension > 1)
	    {
	      score = -search(wToMove^1, 1, -beta, -alpha, depth+extension-1, 
			      pos, plyData, pv, hash, others, tc, true, 0, 
			      centiExtension, PVNode, followPV, thePV, 
			      theMove);
	      ASSERT(score >= -mateScore && score <= mateScore);
	    }
	  else 
	    {
	      score = -qsearch(wToMove^1, 1, 0, -beta, -alpha, pos, 
			       plyData, pv, others, tc, 0, PVNode);
	      ASSERT(score >= -mateScore && score <= mateScore);
	    }

          if (pos->Flags & OneMask[4])
            {
              unmakeMove(wToMove, 1, theMove, pos, plyData);
	      if (pv[0].length == 0)
                {
                  pv[0].length = 1;
                  pv[0].move[0] = firstMove;
                }
              return 0;  /* doesn't matter, not used in main() */
            }
          if (searchMode == 2) score = score / 2;
        }
      
      if (score > alpha)
	{
	  /* this is special because pv[0][0] is theMove in main() */
	  /* update the PV */
	  pv[0].move[0] = theMove;
	  if (!searchMode) pv[0].length = 1;
	  else
	    {
	      ASSERT(pv[1].length <= maxPly)
		for (j = 1; j < pv[1].length; j++)
		  pv[0].move[j] = pv[1].move[j];
	      pv[0].length = pv[1].length;
	    }

	  /* Rearrange move list for  next iteration */
	  for (j = i; j >= 1; j--)
	    allMoves[j] = allMoves[j-1];
	  allMoves[0] = theMove;

#ifdef TREEDUMP
	  printf("Best move so far: %s, score: %d, pvLength: %d\n", 
		 s, score, pv[0].length);	      
#endif
	  
	  if (score >= beta)
	    {
	      unmakeMove(wToMove, 1, theMove, pos, plyData);
#ifdef TREEDUMP
	      printf("%d:%d>=%d (fail high in rsearch())\n", 
		     wToMove, score, beta);
#endif
	      /* killer move */
	      if (score(theMove) <= 1)   /* include even/losing capture */
		{
		  pos->killerFrom[0] = from(theMove);
		  pos->killerTo[0] = to(theMove);
		}
	      return beta;
	    }
	  
	  alpha = score;
	}
      unmakeMove(wToMove, 1, theMove, pos, plyData);
    }

#ifdef TREEDUMP
  printf("%d:%d (fail low in rsearch())\n", wToMove, alpha);
#endif

  ASSERT(firstMove > 0);
  /* In case rsearch fail low */
  if (pv[0].length == 0)
    {
      pv[0].length = 1;
      pv[0].move[0] = firstMove;
    }

  if (movesSearched) return alpha; 
  else
    { 
#ifdef TREEDUMP
      printf("%d: stalemate\n", wToMove);
#endif
      return 0;
    }
}
